using System.Security;
using Autofac;
using FitConnect.Encryption;
using FitConnect.Interfaces.Subscriber;
using FitConnect.Models;
using FitConnect.Models.v1.Api;
using FitConnect.Services.Models;
using FitConnect.Services.Models.v1.Submission;
using IdentityModel;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using NJsonSchema;
using Metadata = FitConnect.Models.Api.Metadata.Metadata;
using SecurityEventToken = FitConnect.Models.SecurityEventToken;

namespace FitConnect;

/// <summary>
///     Fluent API for the FitConnect.Subscriber
/// </summary>
public class Subscriber : FitConnectClient,
    ISubscriber,
    ISubscriberWithSubmission {
    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        string publicKeyEncryption,
        string publicKeySignatureVerification,
        IContainer container) : base(environment, clientId, clientSecret,
        container,
        privateKeyDecryption, privateKeySigning, publicKeyEncryption,
        publicKeySignatureVerification) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, publicKeyEncryption,
            publicKeySignatureVerification,
            container.Resolve<ILogger>());
    }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        string publicKeyEncryption,
        string publicKeySignatureVerification,
        ILogger? logger = null) : base(environment, clientId, clientSecret, logger,
        privateKeyDecryption, privateKeySigning) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, publicKeyEncryption,
            publicKeySignatureVerification, logger);
    }


    /// <summary>
    ///     Receives the available submissions from the server
    /// </summary>
    /// <param name="destinationId"></param>
    /// <param name="skip"></param>
    /// <param name="take"></param>
    /// <returns></returns>
    public IEnumerable<SubmissionForPickupDto> GetAvailableSubmissions(string? destinationId = null,
        int skip = 0,
        int take = 100) {
        var submissionsResult = SubmissionService.ListSubmissions(destinationId, 0, 100);

        // Creating a dictionary of destinationId to submissionIds from the REST API result
        return submissionsResult.Submissions ?? new List<SubmissionForPickupDto>();
    }


    /// <summary>
    ///     Receives a specific submission from the server.
    /// </summary>
    /// <param name="submissionId"></param>
    /// <returns></returns>
    public ISubscriberWithSubmission RequestSubmission(string submissionId) {
        var submission = (Submission)SubmissionService.GetSubmission(submissionId);
        // SuccessCriteria:2.2, 2.3
        var submitEvent = CheckSecurityEventTokens(submission, out var metadataSignature);


        // SuccessCriteria:3.1
        if (metadataSignature != submission.EncryptedMetadata?.Split('.').Last()) {
            var problem = new Problems(Problems.ProblemTypeEnum.IncorrectAuthenticationTag,
                Problems.TitleAuthenticationTagInvalid,
                detail: Problems.DetailAuthenticationMetadataInvalid,
                problemInstance: Problems.ProblemInstanceEnum.Metadata);
            RejectSubmission(submission, problem);
            throw new SecurityEventException(problem);
        }

        string? metadataString;
        try {
            var (metaDataString, _, metaHash) = Encryption.Decrypt(submission.EncryptedMetadata!);
            metadataString = metaDataString;

            if (metadataString == null)
                throw new Exception("Unable to decrypt metadata");
        }
        catch (Exception e) {
            // SuccessCriteria:3.2
            var problem = new Problems(Problems.ProblemTypeEnum.EncryptionIssue,
                Problems.DetailEncryptionIssue);
            RejectSubmission(submission, problem);
            throw new SecurityEventException(problem, e);
        }

        // SuccessCriteria:3.4
        submission.Metadata = JsonConvert.DeserializeObject<Metadata>(metadataString);
        if (submission.Metadata?.Schema == null) {
            var problem = new Problems(Problems.ProblemTypeEnum.MissingSchema,
                Problems.TitleMissingSchema,
                detail: Problems.DetailMissingSchema,
                problemInstance: Problems.ProblemInstanceEnum.Metadata);
            RejectSubmission(problem);
            throw new SecurityEventException(problem);
        }

        JsonSchema? schema;
        try {
            schema = JsonSchema.FromUrlAsync(submission.Metadata?.Schema).Result;
        }
        catch (Exception e) {
            // SuccessCriteria:3.5
            var problem = new Problems(Problems.ProblemTypeEnum.UnsupportedSchema,
                Problems.DetailUnsupportedSchema);
            RejectSubmission(problem);
            throw new SecurityEventException(problem, e);
        }

        VerifyMetadata(submission, metadataString, schema);

        // SuccessCriteria:3.3
        if (submission.Metadata == null) {
            var problem = new Problems(Problems.ProblemTypeEnum.SyntaxViolation,
                Problems.DetailSyntaxViolation);
            RejectSubmission(submission, problem);
            throw new SecurityEventException(problem);
        }

        // SuccessCriteria:3.8
        if (submission.Metadata.PublicServiceType.Identifier != submission.ServiceType.Identifier) {
            var problem = new Problems(Problems.ProblemTypeEnum.ServiceMismatch,
                Problems.DetailServiceMismatch);
            RejectSubmission(problem);
            throw new SecurityEventException(problem);
        }

        // SuccessCriteria:3.9
        var destination = DestinationService.GetDestinationAsync(submission.DestinationId).Result;
        if (destination == null) {
            throw new ArgumentException("Destination not found");
        }

        if (!destination.Services!.Select(s => s.Identifier)
                .Contains(submission.Metadata.PublicServiceType.Identifier)) {
            var problem = new Problems(Problems.ProblemTypeEnum.UnsupportedService,
                Problems.DetailUnsupportedService);
            RejectSubmission(problem);
            throw new SecurityEventException(problem);
        }


        if (submission.EncryptedData != null) {
            var (dataString, _, dataHash) = Encryption.Decrypt(submission.EncryptedData);
            submission.Data = dataString;
            VerifyDataHash(submission, dataString);
        }

        Dictionary<string, string> attachmentSignatures =
            (submitEvent.Payload as dynamic)?.authenticationTags?.attachments
            .ToObject<Dictionary<string, string>>()!;

        submission.Attachments = DownloadAttachments(submission);
        // SuccessCriteria:2.4
        CheckAttachments(submission, attachmentSignatures);

        Submission = submission;
        return this;
    }

    #region Check Submission and Reject if needed

    private void VerifyDataHash(Submission submission, string dataString) {
        if (submission.Metadata?.ContentStructure.Data.Hash.Content ==
            FitEncryption.CalculateHash(dataString)) return;

        Logger?.LogWarning("Data hash mismatch: {DataHash} != {CalculatedHash}",
            submission.Metadata?.ContentStructure.Data.Hash.Content,
            FitEncryption.CalculateHash(dataString));

        // TODO: Check if problem is correct
        RejectSubmission(submission, Problems.EncryptionIssue);
        throw new Exception("Data hash mismatch");
    }


    // SuccessCriteria:3
    private void VerifyMetadata(Submission submission, string metaDataString, JsonSchema schema) {
        var valid = JsonHelper.VerifyMetadata(metaDataString, schema);
        if (!valid) {
            Logger?.LogWarning("Invalid metadata: {MetaData}", metaDataString);

            // SuccessCriteria:3.6
            RejectSubmission(submission,
                new Problems(Problems.ProblemTypeEnum.UnsupportedSchema,
                    Problems.DetailUnsupportedSchema));
            throw new Exception("Metadata validation failed");
        }
    }


    private SecurityEventToken CheckSecurityEventTokens(Submission submission,
        out string metadataSignature) {
        List<SecurityEventToken> status;
        try {
            status = GetStatusForSubmission(submission);
        }
        catch (Exception? e) {
            Logger?.LogWarning("Could not get status for submission: {SubmissionId}",
                submission.Id);
            RejectSubmission(submission,
                new Problems(Problems.ProblemTypeEnum.InvalidEventLog,
                    Problems.DetailEventLogInconsistent));

            throw new SecurityEventException(Problems.TitleEventLogInconsistent,
                Problems.DetailEventLogInconsistent, e);
        }

        if (status.Count(set => set.EventType == EventType.Submit) != 1) {
            RejectSubmission(submission,
                new Problems(Problems.ProblemTypeEnum.InvalidEventLog,
                    Problems.DetailEventLogNotExactlyOneSubmit));
            throw new SecurityEventException(Problems.TitleEventLogInconsistent,
                Problems.DetailEventLogNotExactlyOneSubmit);
        }

        var submitEvent = status.First(set => set.EventType == EventType.Submit);
        var authenticationTag = (submitEvent.Payload as dynamic)?.authenticationTags;

        var dataSignature = authenticationTag?.data?.ToString();
        metadataSignature = authenticationTag?.metadata?.ToString() ?? "";

        if (submission.EncryptedData?.Split('.').Last() != dataSignature) {
            RejectSubmission(submission,
                new Problems(Problems.ProblemTypeEnum.IncorrectAuthenticationTag,
                    Problems.DetailAuthenticationTagMissing));
            throw new AggregateException("Data signature mismatch");
        }

        if (submission.EncryptedMetadata?.Split('.').Last() != metadataSignature) {
            RejectSubmission(submission, Problems.IncorrectAuthenticationTag);
            throw new AggregateException("Metadata signature mismatch");
        }

        return submitEvent;
    }

    private void CheckAttachments(Submission submission,
        Dictionary<string, string> attachmentSignatures) {
        if (submission?.Attachments != null) {
            if (submission.Attachments.Count != attachmentSignatures.Count) {
                RejectSubmission(submission, Problems.AttachmentsMismatch);
                throw new ArgumentException("Attachment count mismatch");
            }

            foreach (var attachment in submission.Attachments) {
                if (attachmentSignatures?[attachment.Id] != attachment.AttachmentAuthentication) {
                    RejectSubmission(submission, Problems.IncorrectAuthenticationTag);
                    throw new AggregateException("Attachment signature mismatch");
                }
            }
        }
    }

    #endregion

    public Submission? Submission { get; private set; }


    /// <summary>
    ///     Reading attachments for a submission.
    /// </summary>
    /// <returns></returns>
    public IEnumerable<Attachment> GetAttachments() {
        if (Submission?.Id == null || Submission?.Metadata == null)
            throw new Exception("No submission available");

        return Submission.Attachments;
    }

    private List<Attachment> DownloadAttachments(Submission submission) {
        var attachments = new List<Attachment>();
        foreach (var id in submission.AttachmentIds) {
            var encryptedAttachment = SubmissionService.GetAttachment(submission.Id, id);
            var (_, content, hash) = Encryption.Decrypt(encryptedAttachment);
            var attachmentMeta =
                submission.Metadata?.ContentStructure.Attachments.First(a => a.AttachmentId == id);

            if (attachmentMeta != null)
                attachments.Add(new Attachment(id, attachmentMeta, content,
                    encryptedAttachment.Split('.').Last()));
        }

        return attachments;
    }

    public void AcceptSubmission() {
        CompleteSubmission(Submission!, FinishSubmissionStatus.Accepted);
    }

    public void RejectSubmission(params Problems[] problems) {
        CompleteSubmission(Submission!, FinishSubmissionStatus.Rejected, problems);
    }

    private void RejectSubmission(Submission submission, params Problems[] problems) {
        CompleteSubmission(submission, FinishSubmissionStatus.Rejected, problems);
    }


    // public void ForwardSubmission() {
    //     CompleteSubmission(Submission!, FinishSubmissionStatus.Forwarded);
    // }

    public void CompleteSubmission(FinishSubmissionStatus status) {
        CompleteSubmission(Submission!, status);
    }

    public bool VerifyStatus(AcceptanceStatus acceptanceStatus) {
        if (Submission == null)
            throw new NullReferenceException("Submission is null");
        var result = true;
        result &= acceptanceStatus.Metadata == Submission.MetaAuthentication;
        result &= acceptanceStatus.Data == Submission.DataAuthentication;
        foreach (var attachment in Submission.Attachments)
            result &= acceptanceStatus.Attachments[attachment.Id] ==
                      attachment.AttachmentAuthentication;

        return result;
    }


    private void CompleteSubmission(Submission submission,
        FinishSubmissionStatus status, Problems[]? problems = null) {
        if (submission.Id == null || submission.CaseId == null ||
            submission.DestinationId == null)
            throw new ArgumentException("Submission does not contain all required fields");

        if (status != FinishSubmissionStatus.Rejected && problems != null)
            throw new ArgumentException("Problems can only be set for rejected submissions");

        var token = status switch {
            FinishSubmissionStatus.Rejected =>
                Encryption.CreateRejectSecurityEventToken(submission.Id,
                    submission.CaseId,
                    submission.DestinationId, problems),
            FinishSubmissionStatus.Accepted => Encryption.CreateAcceptSecurityEventToken(
                submission),
            _ => throw new ArgumentOutOfRangeException(nameof(status), status, null)
        };

        Logger?.LogDebug("Token to accept submission: {Token}", token);
        var result = CasesService.FinishSubmission(submission.CaseId, token);
        Logger?.LogInformation("Submission completed {Status}", result);
    }

    public static string VerifyCallback(string callbackSecret,
        long timestamp, string body) {
        if (timestamp < DateTime.Now.AddMinutes(SecuritySpecification.MaxCallbackAge * -1)
                .ToEpochTime())
            throw new ArgumentException("Request is too old");
        var hmac = SecuritySpecification.CalculateCallbackHmac(callbackSecret, timestamp, body);
        return Convert.ToHexString(hmac).ToLower();
    }


    public static bool VerifyCallback(string callbackSecret, HttpRequest request) {
        if (!request.Headers.ContainsKey("callback-timestamp"))
            throw new ArgumentException("Missing callback-timestamp header");

        var timeStampString = request.Headers["callback-timestamp"].ToString();
        if (!long.TryParse(timeStampString, out var timestamp))
            throw new ArgumentException("Invalid callback-timestamp header");


        var authentication = request.Headers["callback-authentication"];
        using var requestStream = request.Body;
        var content = new StreamReader(requestStream).ReadToEnd().Trim();

        var result = VerifyCallback(callbackSecret, timestamp, content);
        if (result != authentication)
            throw new ArgumentException("Request is not authentic");
        return true;
    }
}

public enum FinishSubmissionStatus {
    Accepted,

    Rejected
    // Forwarded
}

public class SecurityEventException : Exception {
    public string Title { get; }
    public string Detail { get; }

    public SecurityEventException(Problems problem, Exception? innerException = null) : base(
        $"{problem.title}: {problem.detail}", innerException) {
        Title = problem.title;
        Detail = problem.detail;
    }

    public SecurityEventException(string title, string detail, Exception? innerException = null) :
        base(
            $"{title}: {detail}", innerException) {
        Title = title;
        Detail = detail;
    }
}
